"""
DAGVIZ provides a simple visualization of Directed Acyclic Graphs.
"""
try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata  # type: ignore

__version__ = importlib_metadata.version(__name__)  # type: ignore

from typing import Any, Callable, Sequence, Union

import networkx as nx
from networkx.algorithms import dag

from .abstract import AbstractPlot
from .render import render
from .istyle import iStyle
from .style import metro
from .dagre import Dagre  # noqa


def make_abstract_plot(
    G: nx.Graph,
    *,
    order: Union[Sequence[Any], Callable[..., Sequence[Any]]] = dag.topological_sort,
) -> AbstractPlot:
    """Generate an abstract plot for a DAG.

    The abstract plot comprises a sequence of rows with varying number of
    columns, where each column provide information on what it contains.

    Args:
        G:     The DAG to be visualized
        order: Optional; The order of the nodes, or a function that creates an order
               from the graph. The default is to use a topological sort of the nodes.

    Returns:
        An abstract plot of the graph that can be used to render an image from.
    """
    plot = AbstractPlot()
    if isinstance(order, Sequence):
        sequence = order
    else:
        sequence = order(G)
    for nd in sequence:
        row = plot.add_row(G.nodes[nd].get("label", f"{nd}"))
        for pred in G.pred[nd]:
            row.add_input(pred)

        row.add_node(nd, len(G.succ[nd]))

    return plot


def render_svg(
    G: nx.Graph, *, style: Callable[..., iStyle] = metro.svg_renderer()
) -> str:
    """
    Generate a DAG visualization as an SVG string.

    Args:
        G:     The DAG to visualize
        style: Optional; The visualization style to apply
    Returns:
        A string containing the SVG of the plot
    """
    return render(make_abstract_plot(G), style)


class Metro:
    """
    Render a topological ordering of a DAG using the "metro" style in
    jupyter notebooks.

    Args:
        G: directed acyclic graph to render
    """

    def __init__(self, G: nx.DiGraph):
        self.graph = G

    def _repr_html_(self) -> str:
        return render_svg(self.graph)



"""
Rendering of an abstract plot to a concrete format.
"""
from typing import Callable, List

from .abstract import AbstractColumn, AbstractPlot
from .istyle import iStyle


def render(plot: AbstractPlot, styler: Callable[..., iStyle]) -> str:
    """
    Render the plot as an SVG using the specified style.

    Args:
        plot: The abstract plot to render
        styler: The style to use in the rendering. For example `dagviz.style.metro.svg_renderer`
    Returns:
        A string with the SVG content.
    """
    style: iStyle = styler(plot.colors, -plot.columns.start)
    for row in plot.rows:
        last_col = 0
        arcs: List[AbstractColumn] = []
        nodepos = (0, 0)
        node_pos = 1
        for col in row:
            last_col = col.column
            curpos = (col.column, row.row)
            if col.is_node:
                nodepos = curpos
                style.place_node(curpos, col.color)
                if len(arcs) != 0:
                    style.place_left_hline(
                        (arcs[0].column, row.row), curpos, arcs[0].color
                    )
                    arcs = []
                arcs.append(col)

                node_pos = 0
            if col.is_input:
                if node_pos < 0:
                    arcs.append(col)
                    style.place_left_arc(curpos, col.color)
                    if col.is_last:
                        style.place_vline_arc(
                            (col.column, col.start_row.row),
                            curpos,
                            col.start_row.columns[col.column].color,
                        )

                elif node_pos == 0:
                    style.place_vline_node(
                        (col.column, col.start_row.row),
                        curpos,
                        col.start_row.columns[col.column].color,
                    )
                else:
                    arcs.append(col)
                    style.place_right_arc(curpos, col.color)
                    if col.is_last:
                        style.place_vline_arc(
                            (col.column, col.start_row.row),
                            curpos,
                            col.start_row.columns[col.column].color,
                        )
            if col.is_node:
                node_pos = -1
        if len(arcs) >= 2:
            style.place_right_hline(
                (arcs[0].column, row.row), (arcs[-1].column, row.row), arcs[-1].color
            )

        style.place_label(nodepos, (last_col + 1, row.row), row.label)

    return style.dumps()



"""
Render directed graphs in a jupyter notebook using [dagre-d3](https://github.com/dagrejs/dagre-d3).

This is as simple as:

```py
from dagviz.dagre import Dagre
import networkx as nx

g = nx.DiGraph()
g.add_edge("a","b")
Dagre(g)
```
"""
import networkx as nx
from string import Template
import json

_css_text = """
text {
  font-weight: 300;
  font-family: sans-serif;
  font-size: 14px;
}

.node rect {
  stroke: #999;
  fill: #fff;
  stroke-width: 1.5px;
}

.edgePath path {
  stroke: #333;
  stroke-width: 1.5px;
}
"""

_template = """
element.html('<style id="css">$css</style>');
require.config({
    paths: {
        d3: 'https://d3js.org/d3.v5.min',
        dagreD3: 'https://cdnjs.cloudflare.com/ajax/libs/dagre-d3/0.6.4/dagre-d3.min'
    }
});
require(["d3", "dagreD3"], function(d3, dag) {
    $buildGraph
    let elem = d3.select(element.get(0))
    let svg = elem.append("svg");
    let inner = svg.append("g")
    let render = new dag.render();
    render(inner, g);
    // Set up zoom support
    var zoom = d3.zoom().on("zoom", function() {
        inner.attr("transform", d3.event.transform);
    });
    svg.call(zoom);
    // Realign the graph, edges may have x<0
    svg.call(zoom.transform, d3.zoomIdentity.translate(g.graph().width/4.0, 20).scale(1));
    svg.attr("height", g.graph().height + 40);
    svg.attr("width", g.graph().width*1.5 + 40);
});
"""


class Dagre:
    """
    Directed graph visualizer for jupyter notebooks using
    [dagre-d3](https://github.com/dagrejs/dagre-d3).

    Args:
        G: directed graph to render
    """

    def __init__(self, G: nx.DiGraph):
        self._css = _css_text
        self.graph = G

    def _repr_javascript_(self) -> str:
        builder = ["let g = new dag.graphlib.Graph().setGraph({});"]
        for n in self.graph.nodes:
            nstr = json.dumps(n)
            label = json.dumps(str(self.graph.nodes[n].get("label", n)))
            builder.append(f'g.setNode({nstr}, {{ "label": {label}}});')
        for e in self.graph.edges.data():
            label = json.dumps(str(e[2].get("label", "")))
            s = json.dumps(e[0])
            d = json.dumps(e[1])
            builder.append(f'g.setEdge({s},{d}, {{"label": {label}}});')

        return Template(_template).substitute(
            css=self._css.replace("\n", ""), buildGraph="\n".join(builder)
        )



"""
Color palette generation.
"""
from colorsys import hls_to_rgb
from typing import List, Sequence


def palette(ncolors: int) -> Sequence[str]:
    """
    Generate a sequence of colors. The colors should be distinct, but this
    function only does a best effort.

    Args:
        ncolors: the number of colors to generate.
    """
    colors: List[str] = []
    hue, lightness, saturation = 0.0, 1.0, 0.5
    for i in range(ncolors):
        r, g, b = hls_to_rgb(hue / 6, 0.4 + (lightness / 2) * 0.3, saturation)
        hue = (hue + 1) % 7
        if hue == 0:
            lightness = (lightness + 1) % 3
        colors.append(f"#{int(r*256):02x}{int(g*256):02x}{int(b*256):02x}")
    return colors



"""
Metro is the default, and currently only, style for drawing DAGs.
"""
from dagviz.istyle import iStyle
import math
from dataclasses import dataclass
from typing import Any, Callable, Optional, Sequence, Tuple

import svgwrite as svg

from .colors import palette

_XY = Tuple[float, float]


def _arc(
    dwg: svg.Drawing,
    p0: _XY,
    p1: _XY,
    radius: float,
    clockwise: bool = True,
    **kwargs: Any,
) -> Any:
    """ Adds an arc that bulges to the right as it moves from p0 to p1 """
    x0 = p0[0]
    y0 = p0[1]
    x1 = p1[0] - p0[0]
    y1 = p1[1] - p0[1]
    sweep = 1 if clockwise else 0
    ellipseRotation = 0  # has no effect for circles
    return dwg.path(
        d=f"M {x0},{y0} a{radius},{radius} {ellipseRotation} 0,{sweep} {x1},{y1}",
        **kwargs,
    )


@dataclass
class StyleConfig:
    """
    Styling options for the metro style.
    """

    scale: float = 10.0
    "The size of a single cell."

    node_radius: float = 6.0
    "The size of a node."

    node_fill: Optional[str] = None
    "Optional. The fill color of the node. Colors will be assigned automatically if unspecified."

    node_stroke: str = "white"
    "The border color of a node"

    node_stroke_width: float = 2.0
    "The border width of a node"

    edge_stroke_width: float = 2.0
    "The line width of an edge"

    label_font_family: str = "sans-serif"
    "The font family of the label"

    label_arrow_stroke: str = "lightgrey"
    "The line color for the line from the label to the node"

    label_arrow_dash_array: str = "2"
    "The dashing style for the line from the label to the node"

    arc_radius: float = 15.0
    "The radius of an input arc"


class _Style(iStyle):
    d: svg.Drawing
    background: svg.container.Group
    edges: svg.container.Group
    nodes: svg.container.Group
    shift: int
    colors: Sequence[str]
    box_top: float
    box_left: float
    box_bottom: float
    box_right: float

    def __init__(
        self,
        config: StyleConfig,
        colors: int,
        shift: int = 0,
    ):
        self.d = svg.Drawing()
        self.background = self.d.g()
        self.vlines = self.d.g()
        self.hline_borders = self.d.g()
        self.hlines = self.d.g()
        self.nodes = self.d.g()
        self.d.add(self.background)
        self.d.add(self.vlines)
        self.d.add(self.hline_borders)
        self.d.add(self.hlines)
        self.d.add(self.nodes)
        self._top = math.inf
        self._left = math.inf
        self._bottom = -math.inf
        self._right = -math.inf
        self.config = config
        self.shift = shift
        self.colors = palette(colors)

    def dumps(self) -> str:
        t, l, b, r = self._box()
        self.d.viewbox(l, t, r - l, b - t)
        return self.d.tostring()

    def _box(self) -> Tuple[float, float, float, float]:
        return (
            self._top - self.config.scale,
            self._left - self.config.scale,
            self._bottom + self.config.scale,
            max(500, self._right * 2),
        )

    def coord(self, xy: Tuple[int, int], dxy: _XY = (0.0, 0.0)) -> _XY:
        x = self.config.scale * (xy[0] + self.shift) * 2 + self.config.arc_radius * (
            dxy[0] + 1
        )
        y = self.config.scale * xy[1] * 2 + self.config.arc_radius * (dxy[1] + 1)
        self._top = min(y, self._top)
        self._left = min(x, self._left)
        self._bottom = max(y, self._bottom)
        self._right = max(x, self._right)
        return (x, y)

    def top(self, xy: Tuple[int, int]) -> _XY:
        return self.coord(xy, (0, -1))

    def right(self, xy: Tuple[int, int]) -> _XY:
        return self.coord(xy, (1, 0))

    def left(self, xy: Tuple[int, int]) -> _XY:
        return self.coord(xy, (-1, 0))

    def place_node(self, at: Tuple[int, int], color: int) -> None:
        self.nodes.add(
            self.d.circle(
                self.coord(at),
                self.config.node_radius,
                fill=self.config.node_fill or self.colors[color],
                stroke=self.config.node_stroke,
                stroke_width=self.config.node_stroke_width,
            )
        )

    def _place_edge(
        self, layer: svg.container.Group, a: _XY, b: _XY, color: int
    ) -> None:
        layer.add(
            self.d.line(
                a,
                b,
                stroke_width=self.config.edge_stroke_width,
                stroke=self.colors[color],
            )
        )

    def _place_hline_border(self, a: _XY, b: _XY) -> None:
        self.hline_borders.add(
            self.d.line(
                a,
                b,
                stroke_width=self.config.edge_stroke_width
                + 2 * min(self.config.edge_stroke_width, self.config.node_stroke_width),
                stroke="white",
            )
        )

    def place_left_hline(
        self, left: Tuple[int, int], right: Tuple[int, int], color: int
    ) -> None:
        a, b = self.right(left), self.coord(right)
        self._place_hline_border(a, b)
        self._place_edge(self.hlines, a, b, color)

    def place_right_hline(
        self, left: Tuple[int, int], right: Tuple[int, int], color: int
    ) -> None:
        a, b = self.coord(left), self.left(right)
        self._place_hline_border(a, b)
        self._place_edge(self.hlines, a, b, color)

    def place_vline_arc(
        self, top: Tuple[int, int], bottom: Tuple[int, int], color: int
    ) -> None:
        self._place_edge(self.vlines, self.coord(top), self.top(bottom), color)

    def place_vline_node(
        self, top: Tuple[int, int], bottom: Tuple[int, int], color: int
    ) -> None:
        self._place_edge(self.vlines, self.coord(top), self.coord(bottom), color)

    def place_left_arc(self, at: Tuple[int, int], color: int) -> None:
        center = self.coord(at)
        a = (center[0], center[1] - self.config.arc_radius)
        b = (center[0] - self.config.arc_radius, center[1])
        self.hlines.add(
            _arc(
                self.d,
                a,
                b,
                self.config.arc_radius,
                clockwise=True,
                stroke=self.colors[color],
                stroke_width=self.config.edge_stroke_width,
                fill="none",
            )
        )

    def place_right_arc(self, at: Tuple[int, int], color: int) -> None:
        center = self.coord(at)
        a = (center[0], center[1] - self.config.arc_radius)
        b = (center[0] + self.config.arc_radius, center[1])
        self.hlines.add(
            _arc(
                self.d,
                a,
                b,
                self.config.arc_radius,
                clockwise=False,
                stroke=self.colors[color],
                stroke_width=self.config.edge_stroke_width,
                fill="none",
            )
        )

    def place_label(
        self, nodepos: Tuple[int, int], at: Tuple[int, int], label: str
    ) -> None:
        self.nodes.add(
            self.d.text(
                label,
                self.coord(at),
                dominant_baseline="middle",
                font_family=self.config.label_font_family,
            )
        )
        self.background.add(
            self.d.line(
                self.right(nodepos),
                self.coord(at, (-0.4, 0.0)),
                stroke=self.config.label_arrow_stroke,
                stroke_dasharray=self.config.label_arrow_dash_array,
            )
        )


def svg_renderer(config: StyleConfig = StyleConfig()) -> Callable[..., iStyle]:
    """
    Create a renderer with the specified style configuration.

    Args:
        config: the configuration of the metro style
    """

    def builder(colors: int, shift: int = 0) -> iStyle:
        return _Style(config, colors, shift)

    return builder



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dagviz', 'dagviz.style']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.5.1,<3.0.0', 'svgwrite>=1.4.1,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1.0,<2.0']}

setup_kwargs = {
    'name': 'dagviz',
    'version': '0.3.0',
    'description': 'Visualization for DAGs',
    'long_description': '# DAGVIZ\nDAGVIZ offers a "git log"-like visualization for DAGs represented in networkx.\n\n## Installation\n\nDAGVIZ can be installed with `pip`:\n\n```bash\npip install dagviz\n```\n\n## Documentation\n\nDocumentation is available at https://wimyedema.github.io/dagviz/.\n## Usage\n\nDAGVIZ relies on networkx for the representation and manipulation of the\nDAG. An SVG file can be generated as follows:\n\n```py\n    import dagviz\n    import networkx as nx\n    # Create the DAG\n    G = nx.DiGraph()\n    for i in range(5):\n        G.add_node(f"n{i}")\n    G.add_edge(f"n0", f"n1")\n    G.add_edge(f"n0", f"n2")\n    G.add_edge(f"n0", f"n4")\n    G.add_edge(f"n1", f"n3")\n    G.add_edge(f"n2", f"n3")\n    G.add_edge(f"n3", f"n4")\n    # Create an SVG as a string\n    r = dagviz.render_svg(G)\n    with open("simple.svg", "wt") as fs:\n        fs.write(r)\n```\n',
    'author': 'Wim Yedema',
    'author_email': 'wim.yedema@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/WimYedema/dagviz.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Interface for rendering a plot.
"""
from abc import ABC, abstractmethod
from typing import Tuple


class iStyle(ABC):
    """
    Visualization style interface. This interface is used to render the
    various components of the plot. All component coordinates tuples of
    (column, row), and the colors are indices of a palette that must be
    specified by the style.
    """

    @abstractmethod
    def dumps(self) -> str:
        """
        Produce a string for the rendering of the graph.

        Returns:
            a string representing the graph in the appropriate format.
        """
        ...

    @abstractmethod
    def place_node(self, at: Tuple[int, int], color: int) -> None:
        """
        Place a node on the plot.

        Args:
            at:    The (column, row) coordinate of the node
            color: The color number to use for this node
        """
        ...

    @abstractmethod
    def place_left_hline(
        self, left: Tuple[int, int], right: Tuple[int, int], color: int
    ) -> None:
        """
        Place a horizontal line from a node input on to a node, where the
        input is on the left of the node.

        Args:
            left:  The (column, row) coordinate of the node input.
            right: The (column, row) coordinate of the node .
            color: The color number to use for this line.

        Note:
            The rows of *left* and *right* will have the same value.
        """
        ...

    @abstractmethod
    def place_right_hline(
        self, left: Tuple[int, int], right: Tuple[int, int], color: int
    ) -> None:
        """
        Place a horizontal line from a node input on to a node, where the
        input is on the right of the node.

        Args:
            left:  The (column, row) coordinate of the node.
            right: The (column, row) coordinate of the node input.
            color: The color number to use for this line.

        Note:
            The rows of *left* and *right* will have the same value.
        """
        ...

    @abstractmethod
    def place_vline_arc(
        self, top: Tuple[int, int], bottom: Tuple[int, int], color: int
    ) -> None:
        """
        Place a vertical line from a node to a node input, where the input
        is either on the left or the right of the node, but not in the same
        column.

        Args:
            top:    The (column, row) coordinate of the node.
            bottom: The (column, row) coordinate of the node input.
            color:  The color number to use for this line.

        Note:
            The columns of *top* and *bottom* will have the same value.
        """
        ...

    @abstractmethod
    def place_vline_node(
        self, top: Tuple[int, int], bottom: Tuple[int, int], color: int
    ) -> None:
        """
        Place a vertical line from a node to a node input, where the input
        is on the same column as the node.

        Args:
            top:    The (column, row) coordinate of the node.
            bottom: The (column, row) coordinate of the node input.
            color:  The color number to use for this line.

        Note:
            The columns of *top* and *bottom* will have the same value.
        """
        ...

    @abstractmethod
    def place_left_arc(self, at: Tuple[int, int], color: int) -> None:
        """
        Place an arc from the top to the left.

        Args:
            at:     The (column, row) coordinate of the arc.
            color:  The color number to use for this arc.
        """
        ...

    @abstractmethod
    def place_right_arc(self, at: Tuple[int, int], color: int) -> None:
        """
        Place an arc from the top to the right.

        Args:
            at:     The (column, row) coordinate of the arc.
            color:  The color number to use for this arc.
        """
        ...

    @abstractmethod
    def place_label(
        self, nodepos: Tuple[int, int], at: Tuple[int, int], label: str
    ) -> None:
        """
        Place a label for a node.

        Args:
            nodepos: The (column, row) coordinate of the node.
            at:      The (column, row) coordinate of the label.
            label:   The text to display
        """
        ...
